# Copyright 2017 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Configuration for communicating with a solver.

Communicating with a solver---submitting a problem, monitoring its progress, receiving
samples---requires configuration of several parameters such as the selected solver,
its URL, an API token, etc. D-Wave Cloud Client provides multiple options for configuring
those parameters:

* One or more locally saved configuration files.
* Environment variables
* Direct setting of key values in functions

These options can be flexibly used together. The standard use is through
the :func:`~dwave.cloud.client.Client.from_config` classmethod.

Configuration values can be specified in multiple ways, ranked in the following
order (with 1 the highest ranked):

1. Values specified as keyword arguments.
2. Values specified as environment variables.
3. Values specified in the configuration file.
4. Values specified in :class:`~dwave.cloud.client.Client` instance ``defaults``.
5. Values specified in :class:`~dwave.cloud.client.Client` class
   :attr:`~dwave.cloud.client.Client.DEFAULTS`.

Configuration files comply with standard Windows INI-like format,
parsable with Python's :mod:`configparser`. An optional `defaults` section
provides default key-value pairs for all other sections. User-defined key-value
pairs (unrecognized keys) are passed through to the client.

Typically configuration files are created, inspected, and changed using interactive
CLI commands from your system's console, such as :code:`dwave config create` and
:code:`dwave config inspect` (run :code:`dwave --help` for information on CLI options).

Environment variables:

    ``DWAVE_CONFIG_FILE``: Configuration file path.

    ``DWAVE_PROFILE``: Name of profile (section).

    ``DWAVE_API_CLIENT``: API client class. Supported values are ``qpu``, ``sw`` and ``hybrid``.

    ``DWAVE_API_REGION``: API region code.

    ``DWAVE_API_ENDPOINT``: API endpoint URL.

    ``DWAVE_METADATA_API_ENDPOINT``: Metadata API endpoint URL.

    ``DWAVE_API_TOKEN``: API authorization token.

    ``DWAVE_API_SOLVER``: Default solver.

    ``DWAVE_API_PROXY``: URL for proxy connections to D-Wave API.

    ``DWAVE_API_HEADERS``: Optional additional HTTP headers.

Examples:
    The following are typical examples of using :func:`~dwave.cloud.client.Client.from_config`
    to create a configured client.

    This first example initializes :class:`~dwave.cloud.client.Client` from an
    explicitly specified configuration file, "~/jane/my_path_to_config/my_cloud_conf.conf"::

        [defaults]
        token = ABC-123456789123456789123456789

        [first-qpu]
        solver = {"qpu": true}

        [feature]
        endpoint = https://url.of.some.dwavesystem.com/sapi
        token = DEF-987654321987654321987654321
        solver = {"num_qubits__gte": 2000, "max_anneal_schedule_points__gte": 4}

    The example code below creates a client object that connects to a D-Wave QPU,
    using :class:`dwave.cloud.qpu.Client` and the first available online D-Wave system
    at the default API endpoint URL (https://cloud.dwavesys.com/sapi).
    The ``feature`` profile specifies a solver selected based on available features,
    namely we're requesting the first solver that has at least 2000 qubits and the
    anneal schedule can be described with at least 4 points.

    >>> from dwave.cloud import Client
    >>> client = Client.from_config(config_file='~/jane/my_path_to_config/my_cloud_conf.conf')  # doctest: +SKIP
    >>> # code that uses client
    >>> client.close()   # doctest: +SKIP

    This second example auto-detects a configuration file on the local system following the
    user/system configuration paths of :func:`get_configfile_paths`. It passes through
    to the instantiated client an unrecognized key-value pair my_param=`my_value`.

    >>> from dwave.cloud import Client
    >>> client = Client.from_config(my_param="my_value")    # doctest: +SKIP
    >>> # code that uses client
    >>> client.close()      # doctest: +SKIP

    This third example instantiates two clients, for managing both QPU and software
    solvers. Common key-value pairs are taken from the defaults section of a shared
    configuration file::

        [defaults]
        token = ABC-123456789123456789123456789

        [primary-qpu]
        solver = {"qpu": true}

        [sw-solver]
        client = sw
        solver = c4-sw_sample
        endpoint = https://url.of.some.software.resource.com/my_if
        token = DEF-987654321987654321987654321

        [backup-qpu]
        solver = {"qpu": true, "num_qubits__gte": 2000}
        endpoint = https://url.of.some.dwavesystem.com/sapi
        proxy = http://user:pass@myproxy.com:8080/
        token = XYZ-0101010100112341234123412341234

    The example code below creates client objects for two QPU solvers (at the
    same URL but each with its own solver ID and token) and one software solver.

    >>> from dwave.cloud import Client
    >>> client_qpu1 = Client.from_config(profile='primary-qpu')    # doctest: +SKIP
    >>> client_qpu1 = Client.from_config(profile='backup-qpu')    # doctest: +SKIP
    >>> client_sw1 = Client.from_config(profile='sw-solver')   # doctest: +SKIP
    >>> client_qpu1.default_solver   # doctest: +SKIP
    'EXAMPLE_2000Q_SYSTEM_A'
    >>> client_qpu2.endpoint   # doctest: +SKIP
    'https://url.of.some.dwavesystem.com/sapi'
    >>> # code that uses client
    >>> client_qpu1.close() # doctest: +SKIP
    >>> client_qpu2.close() # doctest: +SKIP
    >>> client_sw1.close() # doctest: +SKIP

    This fourth example loads configurations auto-detected in more than one configuration
    file, with the higher priority file (in the current working directory) supplementing
    and overriding values from the lower priority user-local file. After instantiation,
    an endpoint from the default section and client from the profile section is provided
    from the user-local ``/usr/local/share/dwave/dwave.conf`` file::

        [defaults]
        solver = {"qpu": true}

        [dw2000]
        endpoint = https://int.se.dwavesystems.com/sapi
        token = ABC-123456789123456789123456789

    A solver is supplemented from the file in the current working directory, which also
    overrides the token value. ``./dwave.conf`` is the file in the current directory::

        [dw2000]
        token = DEF-987654321987654321987654321

    >>> from dwave.cloud import Client
    >>> client = Client.from_config()  # doctest: +SKIP
    >>> client.default_solver   # doctest: +SKIP
    'EXAMPLE_2000Q_SYSTEM_A'
    >>> client.endpoint  # doctest: +SKIP
    'https://int.se.dwavesystems.com/sapi'
    >>> client.token  # doctest: +SKIP
    'DEF-987654321987654321987654321'
    >>> # code that uses client
    >>> client.close() # doctest: +SKIP

    The next example uses :func:`~dwave.cloud.config.load_config` to load profile values.
    **Most users do not need to use this method.** It loads from the following configuration
    file, dwave_c.conf, located in the current working directory, and specified explicitly::

        [defaults]
        endpoint = https://url.of.some.dwavesystem.com/sapi
        solver = {"qpu": true}

        [dw2000a]
        solver = {"software": true, "name": "EXAMPLE_2000Q"}
        token = ABC-123456789123456789123456789

        [dw2000b]
        solver = {"qpu": true}
        token = DEF-987654321987654321987654321

    This configuration file contains two profiles in addition to the defaults section.
    In the following example code, first no profile is specified, and the first profile
    after the defaults section is loaded with the solver overridden by the environment
    variable. Next, the second profile is selected
    with the explicitly named solver overriding the environment variable setting.

    >>> import dwave.cloud
    >>> import os
    >>> os.environ['DWAVE_API_SOLVER'] = 'EXAMPLE_2000Q_SYSTEM'   # doctest: +SKIP
    >>> dwave.cloud.config.load_config("./dwave_c.conf")   # doctest: +SKIP
    {'client': 'sw',
     'endpoint': 'https://url.of.some.dwavesystem.com/sapi',
     'proxy': None,
     'headers': None,
     'solver': 'EXAMPLE_2000Q_SYSTEM',
     'token': 'ABC-123456789123456789123456789'}
    >>> dc.config.load_config("./dwave_c.conf", profile='dw2000b', solver='Solver3')   # doctest: +SKIP
    {'client': 'qpu',
     'endpoint': 'https://url.of.some.dwavesystem.com/sapi',
     'proxy': None,
     'headers': None,
     'solver': 'Solver3',
     'token': 'DEF-987654321987654321987654321'}

"""

import os
import ast
import logging
import configparser

import homebase

from dwave.cloud.exceptions import ConfigFileReadError, ConfigFileParseError
from dwave.cloud.package_info import __version__, __packagename__

__all__ = ['get_configfile_paths', 'get_configfile_path', 'get_default_configfile_path',
           'load_config_from_files', 'load_profile_from_files', 'load_config']

logger = logging.getLogger(__name__)

CONF_APP = "dwave"
CONF_AUTHOR = "dwavesystem"
CONF_FILENAME = "dwave.conf"

ENV_OPTION_MAP = {
    'DWAVE_API_CLIENT': 'client',
    'DWAVE_API_REGION': 'region',
    'DWAVE_API_ENDPOINT': 'endpoint',
    'DWAVE_API_TOKEN': 'token',
    'DWAVE_API_SOLVER': 'solver',
    'DWAVE_API_PROXY': 'proxy',
    'DWAVE_API_HEADERS': 'headers',
    'DWAVE_METADATA_API_ENDPOINT': 'metadata_api_endpoint',
}
"""Map of environment variable names to config options."""

MUTUALLY_EXCLUSIVE_OPTIONS = {
    'region-endpoint-group': {'endpoint', 'region'},
}
"""List of options to be cleared on *lower level*, per each option set on higher
level, in addition to the option value override (assumed).
"""


def parse_float(s, default=None):
    """Parse value as returned by ConfigParse as float.

    NB: we need this instead of ``ConfigParser.getfloat`` when we're parsing
    values downstream.
    """

    if s is None or s == '':
        return default
    return float(s)


def parse_int(s, default=None):
    """Parse value as returned by ConfigParse as int.

    NB: we need this instead of ``ConfigParser.getint`` when we're parsing
    values downstream.
    """

    if s is None or s == '':
        return default
    if float(s) != int(s):
        raise ValueError
    return int(s)


def parse_boolean(s, default=None):
    """Parse value as returned by ConfigParse as bool.

    Rules::

        0, false, off => False
        empty/None    => default
        otherwise     => True

    NB: we need this instead of ``ConfigParser.getboolean`` when we're parsing
    values downstream, out of ConfigParser context.
    """

    if s is None or s == '':
        return default

    if isinstance(s, str):
        s = s.lower()
        if s in ['false', 'off']:
            s = 'False'
        elif s in ['true', 'on']:
            s = 'True'
        try:
            s = ast.literal_eval(s)
        except:
            raise ValueError('invalid boolean value: {!r}'.format(s))

    return bool(s)


def get_configfile_paths(system=True, user=True, local=True, only_existing=True):
    """Return a list of local configuration file paths.

    Search paths for configuration files on the local system
    are based on homebase_ and depend on operating system; for example, for Linux systems
    these might include ``dwave.conf`` in the current working directory (CWD),
    user-local ``.config/dwave/``, and system-wide ``/etc/dwave/``.

    .. _homebase: https://github.com/dwavesystems/homebase

    Args:
        system (boolean, default=True):
            Search for system-wide configuration files.

        user (boolean, default=True):
            Search for user-local configuration files.

        local (boolean, default=True):
            Search for local configuration files (in CWD).

        only_existing (boolean, default=True):
            Return only paths for files that exist on the local system.

    Returns:
        list[str]:
            List of configuration file paths.

    Examples:
        This example displays all paths to configuration files on a Windows system
        running Python 2.7 and then finds the single existing configuration file.

        >>> from dwave.cloud.config import get_configfile_paths
        >>> # Display paths
        >>> get_configfile_paths(only_existing=False)   # doctest: +SKIP
        ['C:\\ProgramData\\dwavesystem\\dwave\\dwave.conf',
         'C:\\Users\\jane\\AppData\\Local\\dwavesystem\\dwave\\dwave.conf',
         '.\\dwave.conf']
        >>> # Find existing files
        >>> get_configfile_paths()   # doctest: +SKIP
        ['C:\\Users\\jane\\AppData\\Local\\dwavesystem\\dwave\\dwave.conf']

    """

    candidates = []

    # system-wide has the lowest priority, `/etc/dwave/dwave.conf`
    if system:
        candidates.extend(homebase.site_config_dir_list(
            app_author=CONF_AUTHOR, app_name=CONF_APP,
            use_virtualenv=False, create=False))

    # user-local will override it, `~/.config/dwave/dwave.conf`
    if user:
        candidates.append(homebase.user_config_dir(
            app_author=CONF_AUTHOR, app_name=CONF_APP, roaming=False,
            use_virtualenv=False, create=False))

    # highest priority (overrides all): `./dwave.conf`
    if local:
        candidates.append(".")

    paths = [os.path.join(base, CONF_FILENAME) for base in candidates]
    if only_existing:
        paths = list(filter(os.path.exists, paths))

    return paths


def get_configfile_path():
    """Return the highest-priority local configuration file.

    Selects the top-ranked configuration file path from a list of candidates returned
    by :func:`get_configfile_paths()`, or ``None`` if no candidate path exists.

    Returns:
        str:
            Configuration file path.

    Examples:
        This example displays the highest-priority configuration file on a
        Windows system.

        >>> from dwave.cloud import config
        >>> # Display paths
        >>> config.get_configfile_paths(only_existing=False)   # doctest: +SKIP
        ['C:\\ProgramData\\dwavesystem\\dwave\\dwave.conf',
         'C:\\Users\\jane\\AppData\\Local\\dwavesystem\\dwave\\dwave.conf',
         '.\\dwave.conf']
        >>> # Find highest-priority local configuration file
        >>> config.get_configfile_path()   # doctest: +SKIP
        'C:\\Users\\jane\\AppData\\Local\\dwavesystem\\dwave\\dwave.conf'

    """
    paths = get_configfile_paths()
    return paths[-1] if paths else None


def get_default_configfile_path():
    """Return the default configuration-file path.

    Typically returns a user-local configuration file; e.g:
    ``~/.config/dwave/dwave.conf``.

    Returns:
        str:
            Configuration file path.

    Examples:
        This example displays the default configuration file on an Ubuntu Linux
        system.

        >>> from dwave.cloud import config
        >>> # Display paths
        >>> config.get_configfile_paths(only_existing=False)   # doctest: +SKIP
        ['/etc/xdg/xdg-ubuntu/dwave/dwave.conf',
         '/usr/share/upstart/xdg/dwave/dwave.conf',
         '/etc/xdg/dwave/dwave.conf',
         '/home/mary/.config/dwave/dwave.conf',
         './dwave.conf']
        >>> # Find default configuration path
        >>> config.get_default_configfile_path()   # doctest: +SKIP
        '/home/mary/.config/dwave/dwave.conf'

    """
    base = homebase.user_config_dir(
        app_author=CONF_AUTHOR, app_name=CONF_APP, roaming=False,
        use_virtualenv=False, create=False)
    path = os.path.join(base, CONF_FILENAME)
    return path


def get_cache_dir():
    """Return a directory path convenient for storing user-local,
    package-local and version-specific cache data.
    """
    return homebase.user_cache_dir(
        app_name=__packagename__, app_author=CONF_AUTHOR,
        version=__version__, use_virtualenv=False, create=True)


def load_config_from_files(filenames=None):
    """Load D-Wave Cloud Client configuration from a list of files.

    .. note:: This method is not standardly used to set up D-Wave Cloud Client configuration.
        It is recommended you use :meth:`.Client.from_config` or
        :meth:`.config.load_config` instead.

    Configuration files comply with standard Windows INI-like format,
    parsable with Python's :mod:`configparser`. A section called
    ``defaults`` contains default values inherited by other sections.

    Each filename in the list (each configuration file loaded) progressively upgrades
    the final configuration, on a key by key basis, per each section.

    Args:
        filenames (list[str], default=None):
            D-Wave Cloud Client configuration files (paths and names).

            If ``None``, searches for a configuration file named ``dwave.conf``
            in all system-wide configuration directories, in the user-local
            configuration directory, and in the current working directory,
            following the user/system configuration paths of :func:`get_configfile_paths`.

    Returns:
        :obj:`~configparser.ConfigParser`:
            :class:`dict`-like mapping of configuration sections (profiles) to
            mapping of per-profile keys holding values.

    Raises:
        :exc:`~dwave.cloud.exceptions.ConfigFileReadError`:
            Config file specified or detected could not be opened or read.

        :exc:`~dwave.cloud.exceptions.ConfigFileParseError`:
            Config file parse failed.

    Examples:
        This example loads configurations from two files. One contains a default
        section with key/values that are overwritten by any profile section that
        contains that key/value; for example, profile dw2000b in file dwave_b.conf
        overwrites the default URL and client type, which profile dw2000a inherits
        from the defaults section, while profile dw2000a overwrites the API token that
        profile dw2000b inherits.

        The files, which are located in the current working directory, are
        (1) dwave_a.conf::

            [defaults]
            endpoint = https://url.of.some.dwavesystem.com/sapi
            client = qpu
            token = ABC-123456789123456789123456789

            [dw2000a]
            solver = EXAMPLE_2000Q_SYSTEM
            token = DEF-987654321987654321987654321

        and (2) dwave_b.conf::

            [dw2000b]
            endpoint = https://url.of.some.other.dwavesystem.com/sapi
            client = sw
            solver = EXAMPLE_2000Q_SYSTEM

        The following example code loads configuration from both these files, with
        the defined overrides and inheritance.

        .. code:: python

            >>> import dwave.cloud as dc
            >>> import sys
            >>> configuration = dc.config.load_config_from_files(["./dwave_a.conf", "./dwave_b.conf"])   # doctest: +SKIP
            >>> configuration.write(sys.stdout)    # doctest: +SKIP
            [defaults]
            endpoint = https://url.of.some.dwavesystem.com/sapi
            client = qpu
            token = ABC-123456789123456789123456789

            [dw2000a]
            solver = EXAMPLE_2000Q_SYSTEM
            token = DEF-987654321987654321987654321

            [dw2000b]
            endpoint = https://url.of.some.other.dwavesystem.com/sapi
            client = sw
            solver = EXAMPLE_2000Q_SYSTEM

    """
    if filenames is None:
        filenames = get_configfile_paths()

    config = configparser.ConfigParser(default_section="defaults")
    for filename in filenames:
        try:
            filename = os.path.expandvars(os.path.expanduser(filename))
            with open(filename, 'r') as f:
                config.read_file(f, filename)
        except (IOError, OSError):
            raise ConfigFileReadError("Failed to read {!r}".format(filename))
        except configparser.Error:
            raise ConfigFileParseError("Failed to parse {!r}".format(filename))
    return config


def load_profile_from_files(filenames=None, profile=None):
    """Load a profile from a list of D-Wave Cloud Client configuration files.

    .. note:: This method is not standardly used to set up D-Wave Cloud Client configuration.
        It is recommended you use :meth:`.Client.from_config` or
        :meth:`.config.load_config` instead.

    Configuration files comply with standard Windows INI-like format,
    parsable with Python's :mod:`configparser`.

    Each file in the list is progressively searched until the first profile is found.
    This function does not input profile information from environment variables.

    Args:
        filenames (list[str], default=None):
            D-Wave cloud client configuration files (path and name). If ``None``,
            searches for existing configuration files in the standard directories
            of :func:`get_configfile_paths`.

        profile (str, default=None):
            Name of profile to return from reading the configuration from the specified
            configuration file(s). If ``None``, progressively falls back in the
            following order:

            (1) ``profile`` key following ``[defaults]`` section.
            (2) First non-``[defaults]`` section.
            (3) ``[defaults]`` section.

    Returns:
        dict:
            Mapping of configuration keys to values. If no valid config/profile
            is found, returns an empty dict.

    Raises:
        :exc:`~dwave.cloud.exceptions.ConfigFileReadError`:
            Config file specified or detected could not be opened or read.

        :exc:`~dwave.cloud.exceptions.ConfigFileParseError`:
            Config file parse failed.

        :exc:`ValueError`:
            Profile name not found.

    Examples:
        This example loads a profile based on configurations from two files. It
        finds the first profile, dw2000a, in the first file, dwave_a.conf, and adds to
        the values of the defaults section, overwriting the existing client value,
        while ignoring the profile in the second file, dwave_b.conf.

        The files, which are located in the current working directory, are
        (1) dwave_a.conf::

            [defaults]
            endpoint = https://url.of.some.dwavesystem.com/sapi
            client = qpu
            token = ABC-123456789123456789123456789

            [dw2000a]
            client = sw
            solver = EXAMPLE_2000Q_SYSTEM_A
            token = DEF-987654321987654321987654321

        and (2) dwave_b.conf::

            [dw2000b]
            endpoint = https://url.of.some.other.dwavesystem.com/sapi
            client = qpu
            solver = EXAMPLE_2000Q_SYSTEM_B

        The following example code loads profile values from parsing both these files,
        by default loading the first profile encountered or an explicitly specified profile.

        >>> from dwave.cloud import config
        >>> config.load_profile_from_files(["./dwave_a.conf", "./dwave_b.conf"])   # doctest: +SKIP
        {'client': 'sw',
         'endpoint': 'https://url.of.some.dwavesystem.com/sapi',
         'solver': 'EXAMPLE_2000Q_SYSTEM_A',
         'token': 'DEF-987654321987654321987654321'}
        >>> config.load_profile_from_files(["./dwave_a.conf", "./dwave_b.conf"],
        ...                                profile='dw2000b')   # doctest: +SKIP
        {'client': 'qpu',
        'endpoint': 'https://url.of.some.other.dwavesystem.com/sapi',
        'solver': 'EXAMPLE_2000Q_SYSTEM_B',
        'token': 'ABC-123456789123456789123456789'}

    """

    # progressively build config from a file, or a list of auto-detected files
    # raises ConfigFileReadError/ConfigFileParseError on error
    config = load_config_from_files(filenames)

    # determine profile name fallback:
    #  (1) profile key under [defaults],
    #  (2) first non-[defaults] section
    #  (3) [defaults] section
    first_section = next(iter(config.sections() + [None]))
    config_defaults = config.defaults()
    if not profile:
        profile = config_defaults.get('profile', first_section)

    if profile:
        try:
            section = dict(config[profile])
        except KeyError:
            raise ValueError("Config profile {!r} not found".format(profile))
    else:
        # as the very last resort (unspecified profile name and
        # no profiles defined in config), try to use [defaults]
        if config_defaults:
            section = config_defaults
        else:
            section = {}

    return section


def get_default_config():
    # Used only internally by the package
    config = configparser.ConfigParser(default_section="defaults")
    config.read_string(u"""
        [defaults]
        # This section provides default values for some variables.
        # Any of these can be overridden in a specific profile definition below,
        # with an environment variable, and in code.

        # D-Wave solver API endpoint URL defaults to a production endpoint
        #endpoint = https://cloud.dwavesys.com/sapi

        # Default client is the generic base client which works with all
        # available remote solvers/samplers. It can be specialized for the
        # QPU resource (QPU sampler), software, and hybrid samplers.
        # Possible values: `qpu`, `sw`, `hybrid`, `base` (equal to unspecified)
        #client = base

        # Profile name to use if otherwise unspecified.
        # If undefined, the first section below will be used as the default profile.
        #profile = prod

        # Feature-based definition of solver to be used for sampling. If defining
        # the solver in here, make sure that solver is provided on the endpoint used
        # and with the client used (`qpu` or `sw`; `base` can handle any).
        #solver = {"qpu": true, "online": true, "num_qubits__gte": 2000}

        # Proxy URL (including authentication credentials) that shall be used
        # for all requests to D-Wave API endpoint URL.
        #proxy = http://user:pass@proxy.org:8080/
    """)
    return config


def load_config(config_file=None, profile=None, **kwargs):
    """Load D-Wave Cloud Client configuration based on a configuration file.

    Configuration values can be specified in multiple ways, ranked in the following
    order (with 1 the highest ranked):

    1. Values specified as keyword arguments in :func:`load_config()`. These values replace
       values read from a configuration file, and therefore must be **strings**, including float
       values for timeouts, boolean flags, and solver feature
       constraints (a dictionary encoded as JSON).
    2. Values specified as environment variables.
    3. Values specified in the configuration file.
    4. Values specified as :class:`~dwave.cloud.client.Client` instance defaults.
    5. Values specified in :class:`~dwave.cloud.client.Client` class
       :attr:`~dwave.cloud.client.Client.DEFAULTS`.

    Configuration-file format is described in :mod:`dwave.cloud.config`.

    Available configuration-file options are identical to
    :class:`~dwave.cloud.client.Client` constructor argument names.

    If the location of the configuration file is not specified, auto-detection
    searches for existing configuration files in the standard directories
    of :func:`get_configfile_paths`.

    If a configuration file explicitly specified, via an argument or
    environment variable, does not exist or is unreadable, loading fails with
    :exc:`~dwave.cloud.exceptions.ConfigFileReadError`. Loading fails
    with :exc:`~dwave.cloud.exceptions.ConfigFileParseError` if the file is
    readable but invalid as a configuration file.

    Similarly, if a profile explicitly specified, via an argument or
    environment variable, is not present in the loaded configuration, loading fails
    with :exc:`ValueError`. Explicit profile selection also fails if the configuration
    file is not explicitly specified, detected on the system, or defined via
    an environment variable.

    Environment variables: ``DWAVE_CONFIG_FILE``, ``DWAVE_PROFILE``,
    ``DWAVE_API_CLIENT``, ``DWAVE_API_REGION``, ``DWAVE_API_ENDPOINT``,
    ``DWAVE_API_TOKEN``, ``DWAVE_API_SOLVER``, ``DWAVE_API_PROXY``,
    ``DWAVE_API_HEADERS``, ``DWAVE_METADATA_API_ENDPOINT``.

    Environment variables are described in :mod:`dwave.cloud.config`.

    Args:
        config_file (str/[str]/None/False/True, default=None):
            Path to configuration file(s).

            If `None`, the value is taken from `DWAVE_CONFIG_FILE` environment
            variable if defined. If the environment variable is undefined or empty,
            auto-detection searches for existing configuration files in the standard
            directories of :func:`get_configfile_paths`.

            If `False`, loading from file(s) is skipped; if `True`, forces auto-detection
            (regardless of the `DWAVE_CONFIG_FILE` environment variable).

        profile (str, default=None):
            Profile name (name of the profile section in the configuration file).

            If undefined, inferred from `DWAVE_PROFILE` environment variable if
            defined. If the environment variable is undefined or empty, a profile is
            selected in the following order:

            1. From the default section if it includes a profile key.
            2. The first section (after the default section).
            3. If no other section is defined besides `[defaults]`, the defaults
               section is promoted and selected.

        **kwargs (dict, optional):
            :class:`~dwave.cloud.client.Client` constructor arguments.

    Returns:
        dict:
            Mapping of configuration keys to values for the profile (section),
            as read from the configuration file and optionally overridden by
            environment values and specified keyword arguments.

    Raises:
        :exc:`ValueError`:
            Invalid (non-existing) profile name.

        :exc:`~dwave.cloud.exceptions.ConfigFileReadError`:
            Config file specified or detected could not be opened or read.

        :exc:`~dwave.cloud.exceptions.ConfigFileParseError`:
            Config file parse failed.

    Note:
        Prior to 0.8.0, some keyword arguments did not overwrite config
        variables when their value was ``None``. Now we consistently do
        :meth:`dict.update` on the config read from file/env for all ``kwargs``.

    Examples:
        This example loads the configuration from an auto-detected configuration file
        in the home directory of a Windows system user.

        >>> from dwave.cloud import config
        >>> config.load_config()         # doctest: +SKIP
        {'client': 'qpu',
         'endpoint': 'https://url.of.some.dwavesystem.com/sapi',
         'proxy': None,
         'solver': 'EXAMPLE_2000Q_SYSTEM_A',
         'token': 'DEF-987654321987654321987654321',
         'headers': None}
        ... # See which configuration file was loaded
        >>> config.get_configfile_paths()             # doctest: +SKIP
        ['C:\\Users\\jane\\AppData\\Local\\dwavesystem\\dwave\\dwave.conf']

        Additional examples are given in :mod:`dwave.cloud.config`.

    """

    logger.trace("load_config(config_file=%r, profile=%r, kwargs=%r)",
                 config_file, profile, kwargs)

    if profile is None:
        profile = os.getenv("DWAVE_PROFILE")

    if config_file == False:
        # skip loading from file altogether
        section = {}
    elif config_file == True:
        # force auto-detection, disregarding DWAVE_CONFIG_FILE
        section = load_profile_from_files(None, profile)
    else:
        # auto-detect if not specified with arg or env
        if config_file is None:
            # note: both empty and undefined DWAVE_CONFIG_FILE treated as None
            config_file = os.getenv("DWAVE_CONFIG_FILE")

        # handle ''/None/str/[str] for `config_file` (after env)
        filenames = None
        if config_file:
            if isinstance(config_file, str):
                filenames = [config_file]
            else:
                filenames = config_file

        section = load_profile_from_files(filenames, profile)

    logger.trace("config (from files) = %r", section)

    # override with env
    update_config_from_environment(section)
    logger.trace("config (from files+env) = %r", section)

    # override with supplied kwarg options
    update_config(section, kwargs)
    logger.trace("config (from files+env+kwargs) = %r", section)

    return section


def update_config_from_environment(section):
    """Update config profile/``section`` with values from environment variables.

    Supported environment variables are listed as keys in
    :attr:`.ENV_OPTION_MAP`, with corresponding config option names as values.
    """

    envopts = {opt: os.getenv(env) for env, opt in ENV_OPTION_MAP.items()}

    update_config(section, envopts)


def update_config(config: dict, options: dict) -> None:
    """Update `config` inplace with `options`, ignoring None and blank string
    values, clearing mutually exclusive options on different levels.
    """
    # skip null and empty option values
    # TODO: not needed when we switch to structured/typed config (like YAML)
    updates = {k: v for k,v in options.items() if v is not None and v != ''}

    # handle mutually exclusive options first (so it's order invariant)
    for group, optionset in MUTUALLY_EXCLUSIVE_OPTIONS.items():
        if updates.keys() & optionset:
            for excluded in optionset:
                config.pop(excluded, None)

    config.update(updates)



# Copyright 2019 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import unittest
import threading
import concurrent.futures

from dwave.cloud.concurrency import (
    _PriorityOrderedItem,
    _PrioritizedWorkItem,
    _PrioritizingQueue,
    PriorityThreadPoolExecutor,
)


class Test_PrioritizedWorkItem(unittest.TestCase):

    def test_construction(self):
        priority = 2
        future = concurrent.futures.Future()
        fn = lambda: None
        args = (1,)
        kwargs = {'a': 1}

        # without priority
        w = concurrent.futures.thread._WorkItem(future, fn, args, kwargs)
        pw = _PrioritizedWorkItem(w)
        self.assertEqual(pw.priority, sys.maxsize)
        self.assertEqual(pw.item.future, future)
        self.assertEqual(pw.item.fn, fn)
        self.assertEqual(pw.item.args, args)
        self.assertEqual(pw.item.kwargs, kwargs)

        # with priority
        kwargs_pri = kwargs.copy()
        kwargs_pri.update(priority=priority)
        w = concurrent.futures.thread._WorkItem(future, fn, args, kwargs_pri)
        pw = _PrioritizedWorkItem(w)
        self.assertEqual(pw.priority, priority)
        self.assertEqual(pw.item.future, future)
        self.assertEqual(pw.item.fn, fn)
        self.assertEqual(pw.item.args, args)
        self.assertEqual(pw.item.kwargs, kwargs)

    def test_priority_ordering(self):
        w1 = _PrioritizedWorkItem(
            concurrent.futures.thread._WorkItem(None, None, (), dict(priority=1)))
        w2 = _PrioritizedWorkItem(
            concurrent.futures.thread._WorkItem(None, None, (), dict(priority=2)))

        # ordered
        self.assertLess(w1, w2)
        self.assertGreater(w2, w1)

        # always greater that None
        none = _PriorityOrderedItem(None)
        self.assertLess(none, w1)
        self.assertGreater(w1, none)
        self.assertNotEqual(w1, none)


class Test_PrioritizingQueue(unittest.TestCase):

    def test_prioritization(self):
        w1 = concurrent.futures.thread._WorkItem(None, None, (1,), dict(priority=1))
        w2 = _PrioritizedWorkItem(
            concurrent.futures.thread._WorkItem(None, None, (2,), dict(priority=2)))
        w3 = None
        w4 = concurrent.futures.thread._WorkItem(None, None, (3,), {})

        q = _PrioritizingQueue()

        # put different types of items in queue
        q.put(w2)
        q.put(w3)
        q.put(w1)
        q.put(w4)

        # verify order on get
        self.assertEqual(q.get(), None)
        self.assertEqual(q.get().args, (1,))
        self.assertEqual(q.get().item.args, (2,))
        self.assertEqual(q.get().args, (3,))
        self.assertTrue(q.empty())

    def test_double_none_edgecase(self):
        # this can happen on interpreter exit (concurrent.futures.thread._python_exit)
        # thread count adjustment, manual calls to executor.shutdown, etc.
        q = _PrioritizingQueue()
        q.put(None)
        q.put(None)
        self.assertEqual(q.get(), None)
        self.assertEqual(q.get(), None)


class TestPriorityThreadPoolExecutor(unittest.TestCase):

    def test_fallback(self):
        """Without priority specified, it falls back to ThreadPoolExecutor mode."""

        counter = threading.BoundedSemaphore(value=3)

        def worker():
            counter.acquire(blocking=False)

        with PriorityThreadPoolExecutor(max_workers=3) as executor:
            fs = [executor.submit(worker) for _ in range(3)]
            concurrent.futures.wait(fs)

        self.assertFalse(counter.acquire(blocking=False))

        # verify executor shutdown (all threads stopped)
        self.assertFalse(any(t.is_alive() for t in executor._threads))

    def test_prioritization(self):

        # we need a dummy busy bee task to block the executor thread pool
        # until we queue actual jobs -- in order to test priority queue
        # (executor might running the first task as soon as it's enqueued)
        go = threading.Event()

        def bee():
            go.wait()

        # we'll collect results here, ensuring only one worker writes at the time
        results = []
        lock = threading.Lock()

        def worker(val):
            with lock:
                results.append(val)

        with PriorityThreadPoolExecutor(max_workers=1) as executor:
            executor.submit(bee)
            fs = [executor.submit(worker, p, priority=p) for p in [2, 0, 3, 1]]
            go.set()
            concurrent.futures.wait(fs)

        self.assertListEqual(results, [0, 1, 2, 3])

        # verify executor shutdown (all threads stopped)
        self.assertFalse(any(t.is_alive() for t in executor._threads))



# Copyright 2017 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import json
import unittest
from unittest import mock

from click.testing import CliRunner
from parameterized import parameterized

from dwave.cloud.cli import cli
from dwave.cloud.config import load_config
from dwave.cloud.testing import isolated_environ

from tests import config, test_config_path, test_config_profile
from tests.test_mock_solver_loading import solver_object


def touch(path):
    """Implements UNIX `touch`."""
    with open(path, 'a'):
        os.utime(path, None)


class TestConfigCreate(unittest.TestCase):

    @parameterized.expand([
        ("simple", [], dict(token="token")),
        ("full", ["--full"], dict(region="na-west-1", endpoint="endpoint",
                                  token="token", client=None, solver=None)),
    ])
    def test_create(self, name, extra_opts, inputs):
        config_file = 'path/to/dwave.conf'
        profile = 'profile'

        runner = CliRunner(mix_stderr=False)
        with runner.isolated_filesystem():
            result = runner.invoke(cli, [
                'config', 'create', '--config-file', config_file, '--profile', profile
            ] + extra_opts, input='\n'.join('' if v is None else v for v in inputs.values()))
            self.assertEqual(result.exit_code, 0)

            # load and verify config
            with isolated_environ(remove_dwave=True):
                config = load_config(config_file, profile=profile)
                for var, val in inputs.items():
                    self.assertEqual(config.get(var), val)

    @parameterized.expand([
        ("simple", [], dict(config_file=None, token="token")),
        ("full", ["--full"], dict(config_file=None, profile=None, region="na-west-1",
                                  endpoint="endpoint", token="token", client=None, solver=None)),
    ])
    def test_default_flows(self, name, extra_opts, inputs):
        runner = CliRunner(mix_stderr=False)
        with runner.isolated_filesystem():
            with mock.patch("dwave.cloud.config.get_configfile_paths", lambda: ['dwave.conf']):
                with mock.patch("dwave.cloud.utils.input", side_effect=inputs):
                    result = runner.invoke(cli, [
                        'config', 'create'
                    ] + extra_opts, input='\n'.join('' if v is None else v for v in inputs.values()))
                    self.assertEqual(result.exit_code, 0)

                # load and verify config
                with isolated_environ(remove_dwave=True):
                    config = load_config()
                    for var, val in inputs.items():
                        if val:   # skip empty default confirmations
                            self.assertEqual(config.get(var), val)

    @parameterized.expand([
        ("simple", [], dict(token="token")),
        ("full", ["--full"], dict(region="na-west-1", endpoint="endpoint",
                                  token="token", client="base", solver="solver")),
    ])
    def test_update(self, name, extra_opts, inputs):
        config_file = 'dwave.conf'
        profile = 'profile'

        runner = CliRunner(mix_stderr=False)
        with runner.isolated_filesystem():
            # create config
            config_body = '\n'.join(f"{k} = old-{v}" for k,v in inputs.items())
            with open(config_file, 'w') as fp:
                fp.write(f"[{profile}]\n{config_body}")

            # verify config before update
            with isolated_environ(remove_dwave=True):
                config = load_config(config_file=config_file)
                for var, val in inputs.items():
                    self.assertEqual(config.get(var), f"old-{val}")

            # update config
            result = runner.invoke(cli, [
                'config', 'create', '-f', config_file, '-p', profile,
            ] + extra_opts, input='\n'.join('' if v is None else v for v in inputs.values()))
            self.assertEqual(result.exit_code, 0)

            # verify config updated
            with isolated_environ(remove_dwave=True):
                config = load_config(config_file=config_file)
                for var, val in inputs.items():
                    self.assertEqual(config.get(var), val)


class TestCli(unittest.TestCase):

    def test_config_ls(self):
        runner = CliRunner(mix_stderr=False)
        with runner.isolated_filesystem():
            touch('dwave.conf')
            with mock.patch('dwave.cloud.config.homebase.site_config_dir_list',
                            lambda **kw: ['system1', 'system2']):
                with mock.patch('dwave.cloud.config.homebase.user_config_dir',
                                lambda **kw: 'user'):
                    with mock.patch('os.path.exists', lambda *x: True):
                        # test listing of all auto-detected config files
                        result = runner.invoke(cli, [
                            'config', 'ls'
                        ])
                        self.assertEqual(result.output.strip(), '\n'.join([
                            os.path.join('system1', 'dwave.conf'),
                            os.path.join('system2', 'dwave.conf'),
                            os.path.join('user', 'dwave.conf'),
                            os.path.join('.', 'dwave.conf')
                        ]))

                        # test --system
                        result = runner.invoke(cli, [
                            'config', 'ls', '--system'
                        ])
                        self.assertEqual(result.output.strip(), '\n'.join([
                            os.path.join('system1', 'dwave.conf'),
                            os.path.join('system2', 'dwave.conf')
                        ]))

                        # test --user
                        result = runner.invoke(cli, [
                            'config', 'ls', '--user'
                        ])
                        self.assertEqual(result.output.strip(),
                                         os.path.join('user', 'dwave.conf'))

                        # test --local
                        result = runner.invoke(cli, [
                            'config', 'ls', '--local'
                        ])
                        self.assertEqual(result.output.strip(),
                                         os.path.join('.', 'dwave.conf'))

                    # test --include-missing (none of the examined paths exist)
                    with mock.patch('os.path.exists', lambda *x: False):
                        # test listing of all examined paths
                        result = runner.invoke(cli, [
                            'config', 'ls', '--include-missing'
                        ])
                        self.assertEqual(result.output.strip(), '\n'.join([
                            os.path.join('system1', 'dwave.conf'),
                            os.path.join('system2', 'dwave.conf'),
                            os.path.join('user', 'dwave.conf'),
                            os.path.join('.', 'dwave.conf')
                        ]))

                        # test none exists
                        result = runner.invoke(cli, [
                            'config', 'ls', '--system', '--user', '--local'
                        ])
                        self.assertEqual(result.output.strip(), '')

                        # test --system
                        result = runner.invoke(cli, [
                            'config', 'ls', '--system', '--include-missing'
                        ])
                        self.assertEqual(result.output.strip(), '\n'.join([
                            os.path.join('system1', 'dwave.conf'),
                            os.path.join('system2', 'dwave.conf')
                        ]))

                        # test --user
                        result = runner.invoke(cli, [
                            'config', 'ls', '--user', '--include-missing'
                        ])
                        self.assertEqual(result.output.strip(),
                                         os.path.join('user', 'dwave.conf'))

                        # test --local
                        result = runner.invoke(cli, [
                            'config', 'ls', '--local', '--include-missing'
                        ])
                        self.assertEqual(result.output.strip(),
                                         os.path.join('.', 'dwave.conf'))

    def test_configure_inspect(self):
        runner = CliRunner(mix_stderr=False)
        with runner.isolated_filesystem():
            config_file = 'dwave.conf'
            with open(config_file, 'w') as f:
                f.write('''
                    [defaults]
                    endpoint = 1
                    [a]
                    endpoint = 2
                    [b]
                    token = 3''')

            # test auto-detected case
            with mock.patch('dwave.cloud.config.get_configfile_paths',
                            lambda **kw: [config_file]):
                result = runner.invoke(cli, [
                    'config', 'inspect'
                ])
                self.assertIn('endpoint = 2', result.output)

            # test explicit config
            result = runner.invoke(cli, [
                'config', 'inspect', '--config-file', config_file
            ])
            self.assertIn('endpoint = 2', result.output)

            # test explicit profile
            result = runner.invoke(cli, [
                'config', 'inspect', '--config-file', config_file, '--profile', 'b'
            ])
            self.assertIn('endpoint = 1', result.output)
            self.assertIn('token = 3', result.output)

    @parameterized.expand([
        ("--config-file", ),
        ("-f", ),
    ])
    def test_ping(self, config_file_option):
        config_file = 'dwave.conf'
        profile = 'profile'
        client_type = 'qpu'
        params = dict(num_reads=10)
        label = 'label'

        with mock.patch('dwave.cloud.cli.Client') as m:
            # mock returned solver
            client = m.from_config.return_value
            client.get_solver.return_value.nodes = [5, 7, 3]

            runner = CliRunner(mix_stderr=False)
            with runner.isolated_filesystem():
                touch(config_file)
                result = runner.invoke(cli, ['ping',
                                             config_file_option, config_file,
                                             '--profile', profile,
                                             '--client', client_type,
                                             '--sampling-params', json.dumps(params),
                                             '--request-timeout', '.5',
                                             '--polling-timeout', '30',
                                             '--label', label])

            # proper arguments passed to Client.from_config?
            m.from_config.assert_called_with(
                config_file=config_file, profile=profile,
                endpoint=None, region=None,
                client=client_type, solver=None,
                request_timeout=0.5, polling_timeout=30)

            # get solver called?
            client.get_solver.assert_called_with()

            # sampling method called on solver with correct params?
            solver = client.get_solver.return_value
            solver.sample_ising.assert_called_with(
                {3: 0}, {}, label=label, **params)

        self.assertEqual(result.exit_code, 0)

    @parameterized.expand([
        ("--config-file", ),
        ("-f", ),
    ])
    def test_sample(self, config_file_option):
        config_file = 'dwave.conf'
        profile = 'profile'
        client = 'qpu'
        solver = '{"qpu": true}'
        biases = '[0]'
        couplings = '{(0, 4): 1}'
        num_reads = '10'
        label = 'label'

        with mock.patch('dwave.cloud.cli.Client') as m:

            runner = CliRunner(mix_stderr=False)
            with runner.isolated_filesystem():
                touch(config_file)
                result = runner.invoke(cli, ['sample',
                                             config_file_option, config_file,
                                             '--profile', profile,
                                             '--client', client,
                                             '--solver', solver,
                                             '--label', label,
                                             '-h', biases,
                                             '-j', couplings,
                                             '-n', num_reads])

            # proper arguments passed to Client.from_config?
            m.from_config.assert_called_with(
                config_file=config_file, profile=profile,
                endpoint=None, region=None,
                client=client, solver=solver)

            # get solver called?
            c = m.from_config.return_value
            c.get_solver.assert_called_with()

            # sampling method called on solver?
            s = c.get_solver.return_value
            s.sample_ising.assert_called_with(
                {0: 0}, {(0, 4): 1}, num_reads=10, label=label)

        self.assertEqual(result.exit_code, 0)

    @parameterized.expand([
        ("--config-file", ),
        ("-f", ),
    ])
    def test_solvers(self, config_file_option):
        config_file = 'dwave.conf'
        profile = 'profile'
        client_type = 'base'
        solver = '{"qpu": true}'
        solvers = [solver_object('A'), solver_object('B')]

        with mock.patch('dwave.cloud.cli.Client') as m:
            # mock returned solvers
            client = m.from_config.return_value.__enter__.return_value
            client.get_solvers.return_value = solvers

            runner = CliRunner(mix_stderr=False)
            with runner.isolated_filesystem():
                touch(config_file)
                result = runner.invoke(cli, ['solvers',
                                             config_file_option, config_file,
                                             '--profile', profile,
                                             '--client', client_type,
                                             '--solver', solver])

            # proper arguments passed to Client.from_config?
            m.from_config.assert_called_with(
                config_file=config_file, profile=profile,
                endpoint=None, region=None,
                client=client_type, solver=solver)

            # get solvers called?
            client.get_solvers.assert_called_with()

        # verify exit code and stdout printout
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(result.output.count('Solver:'), 2)
        self.assertIn('Solver: A', result.output)
        self.assertIn('Solver: B', result.output)

    @parameterized.expand([
        ("--config-file", ),
        ("-f", ),
    ])
    def test_upload(self, config_file_option):
        config_file = 'dwave.conf'
        profile = 'profile'
        client_type = 'base'
        format = 'dimodbqm'
        problem_id = 'prob:lem:id'
        filename = 'filename'

        with mock.patch('dwave.cloud.cli.Client') as m:

            runner = CliRunner(mix_stderr=False)
            with runner.isolated_filesystem():
                touch(config_file)
                touch(filename)
                result = runner.invoke(cli, ['upload',
                                             config_file_option, config_file,
                                             '--profile', profile,
                                             '--client', client_type,
                                             '--format', format,
                                             '--problem-id', problem_id,
                                             filename])

                # proper arguments passed to Client.from_config?
                m.from_config.assert_called_with(
                    config_file=config_file, profile=profile,
                    endpoint=None, region=None,
                    client=client_type)

                # upload method called on client?
                c = m.from_config.return_value
                self.assertTrue(c.upload_problem_encoded.called)

                # verify problem_id
                args, kwargs = c.upload_problem_encoded.call_args
                self.assertEqual(kwargs.get('problem_id'), problem_id)

        self.assertEqual(result.exit_code, 0)

    def test_platform(self):
        runner = CliRunner(mix_stderr=False)
        result = runner.invoke(cli, ['--platform'])

        # verify exit code and stdout printout
        self.assertEqual(result.exit_code, 0)

        from dwave.cloud.package_info import __packagename__, __version__
        self.assertNotIn(__packagename__, result.output)
        required = ['python', 'machine', 'system', 'platform']
        for key in required:
            self.assertIn(key, result.output)


@unittest.skipUnless(config, "No live server configuration available.")
class TestCliLive(unittest.TestCase):

    def test_ping(self):
        runner = CliRunner(mix_stderr=False)
        result = runner.invoke(cli, ['ping',
                                     '--config-file', test_config_path,
                                     '--profile', test_config_profile])
        self.assertEqual(result.exit_code, 0)

    def test_ping_json(self):
        runner = CliRunner(mix_stderr=False)
        result = runner.invoke(cli, ['ping',
                                     '--config-file', test_config_path,
                                     '--profile', test_config_profile,
                                     '--json'])

        res = json.loads(result.output)
        self.assertIn('timestamp', res)
        self.assertIn('datetime', res)
        self.assertIn('solver_id', res)
        self.assertIn('code', res)
        self.assertEqual(result.exit_code, 0)

    def test_ping_json_timeout_error(self):
        runner = CliRunner(mix_stderr=False)
        result = runner.invoke(cli, ['ping',
                                     '--config-file', test_config_path,
                                     '--profile', test_config_profile,
                                     '--polling-timeout', '0.00001',
                                     '--json'])

        res = json.loads(result.output)
        self.assertIn('timestamp', res)
        self.assertIn('datetime', res)
        self.assertIn('code', res)
        self.assertIn('error', res)
        self.assertEqual(result.exit_code, 9)

    @parameterized.expand([
        ("bqm", 3),
        ("cqm", 5),
        ("dqm", 5),
    ])
    def test_ping_unstructured_solver(self, problem_type, time_limit):
        solver = json.dumps({"supported_problem_types__contains": problem_type})
        params = json.dumps({"time_limit": time_limit})
        runner = CliRunner(mix_stderr=False)
        result = runner.invoke(cli, ['ping',
                                     '--config-file', test_config_path,
                                     '--profile', test_config_profile,
                                     '--solver', solver,
                                     '--sampling-params', params])
        self.assertEqual(result.exit_code, 0)


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid
import unittest
from urllib.parse import urljoin

import requests_mock

from dwave.cloud.api.resources import Solvers
from dwave.cloud.api import exceptions, models
from dwave.cloud.client.base import Client
from dwave.cloud.testing.mocks import qpu_clique_solver_data

from tests import config


class TestMockSolvers(unittest.TestCase):
    """Test request formation and response parsing (including error handling)
    works correctly for all :class:`dwave.cloud.api.resources.Solvers` methods.
    """

    token = str(uuid.uuid4())
    endpoint = 'http://test.com/path/'

    def setUp(self):
        self.mocker = requests_mock.Mocker()

        solver1_data = qpu_clique_solver_data(3)
        solver1_id = solver1_data['id']
        solver1_uri = urljoin(self.endpoint, 'solvers/remote/{}'.format(solver1_id))

        solver2_data = qpu_clique_solver_data(5)
        solver2_id = solver2_data['id']
        solver2_uri = urljoin(self.endpoint, 'solvers/remote/{}'.format(solver2_id))

        all_solver_data = [solver1_data, solver2_data]
        all_solver_data_uri = urljoin(self.endpoint, 'solvers/remote/')
        self.solver_ids = [solver1_id, solver2_id]
        self.solver_sizes = [3, 5]

        headers = {'X-Auth-Token': self.token}

        self.mocker.get(requests_mock.ANY, status_code=401)
        self.mocker.get(requests_mock.ANY, status_code=404, request_headers=headers)

        self.mocker.get(solver1_uri, json=solver1_data, request_headers=headers)
        self.mocker.get(solver2_uri, json=solver2_data, request_headers=headers)
        self.mocker.get(all_solver_data_uri, json=all_solver_data, request_headers=headers)

        self.mocker.start()

    def tearDown(self):
        self.mocker.stop()

    def test_list_solvers(self):
        """List of solver configurations fetched."""

        resource = Solvers(token=self.token, endpoint=self.endpoint)
        solvers = resource.list_solvers()

        self.assertEqual(len(solvers), 2)
        self.assertEqual(solvers[0].properties['num_qubits'], self.solver_sizes[0])
        self.assertEqual(solvers[1].properties['num_qubits'], self.solver_sizes[1])

    def test_get_solver(self):
        """Specific solver config retrieved (by id)."""

        solver_id = self.solver_ids[0]
        num_qubits = self.solver_sizes[0]

        resource = Solvers(token=self.token, endpoint=self.endpoint)
        solver = resource.get_solver(solver_id)

        self.assertEqual(solver.id, solver_id)
        self.assertEqual(solver.properties['num_qubits'], num_qubits)

    def test_nonexisting_solver(self):
        """Not found error is raised when trying to fetch a non-existing solver."""

        resource = Solvers(token=self.token, endpoint=self.endpoint)
        with self.assertRaises(exceptions.ResourceNotFoundError):
            resource.get_solver('non-existing-solver')

    def test_invalid_token(self):
        """Auth error is raised when request not authorized with token."""

        resource = Solvers(token='invalid-token', endpoint=self.endpoint)
        with self.assertRaises(exceptions.ResourceAuthenticationError):
            resource.list_solvers()


@unittest.skipUnless(config, "SAPI access not configured.")
class TestCloudSolvers(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        with Client(**config) as client:
            cls.api = Solvers.from_client_config(client)

    @classmethod
    def tearDownClass(cls):
        cls.api.close()

    def test_list_solvers(self):
        """List of all available solvers retrieved."""

        solvers = self.api.list_solvers()

        self.assertIsInstance(solvers, list)
        self.assertGreater(len(solvers), 0)

        for solver in solvers:
            self.assertIsInstance(solver, models.SolverConfiguration)

    def test_get_solver(self):
        """Specific solver config retrieved (by id)."""

        # don't assume solver availability, instead try fetching one from the list
        solvers = self.api.list_solvers()
        solver_id = solvers.pop().id

        solver = self.api.get_solver(solver_id)
        self.assertIsInstance(solver, models.SolverConfiguration)
        self.assertEqual(solver.id, solver_id)

    def test_nonexisting_solver(self):
        """Not found error is raised when trying to fetch a non-existing solver."""

        with self.assertRaises(exceptions.ResourceNotFoundError):
            self.api.get_solver('non-existing-solver')



# Copyright 2017 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import uuid
import unittest
from unittest import mock

import networkx as nx

try:
    import dimod
except ImportError:
    dimod = None

from dwave.cloud.testing import isolated_environ, iterable_mock_open, mocks


@mock.patch.dict(os.environ)
class TestEnvUtils(unittest.TestCase):
    def setUp(self):
        self.var = str(uuid.uuid4())
        self.val = str(uuid.uuid4())
        self.alt = str(uuid.uuid4())

    def clear(self):
        if self.var in os.environ:
            del os.environ[self.var]

    def setval(self):
        os.environ[self.var] = self.val

    def test_isolation(self):
        # set var to val in "global" env
        self.setval()
        self.assertEqual(os.getenv(self.var), self.val)

        # ensure isolated env stores modified value
        def test():
            os.environ[self.var] = self.alt
            self.assertEqual(os.getenv(self.var), self.alt)

        with isolated_environ():
            test()

        isolated_environ()(test)()

        # ensure "global" env is restored
        self.assertEqual(os.getenv(self.var), self.val)

    def test_update(self):
        # set var to val in "global" env
        self.setval()
        self.assertEqual(os.getenv(self.var), self.val)

        # ensure isolated env has modified value
        args = ({self.var: self.alt}, )

        def test():
            self.assertEqual(os.getenv(self.var), self.alt)

        with isolated_environ(*args):
            test()

        isolated_environ(*args)(test)()

        # ensure "global" env is restored
        self.assertEqual(os.getenv(self.var), self.val)

    def test_add(self):
        # clear var in "global" env
        self.clear()
        self.assertEqual(os.getenv(self.var), None)

        # ensure isolated env adds var
        args = ({self.var: self.val}, )

        def test():
            self.assertEqual(os.getenv(self.var), self.val)

        with isolated_environ(*args):
            test()

        isolated_environ(*args)(test)()

        # ensure "global" env is restored
        self.assertEqual(os.getenv(self.var), None)

    def test_remove(self):
        # set var to val in "global" env
        self.setval()
        self.assertEqual(os.getenv(self.var), self.val)

        # ensure isolated env removes var
        with isolated_environ(remove={self.var: self.val}):
            self.assertEqual(os.getenv(self.var), None)
        with isolated_environ(remove=[self.var]):
            self.assertEqual(os.getenv(self.var), None)
        with isolated_environ(remove=set([self.var])):
            self.assertEqual(os.getenv(self.var), None)

        # ensure "global" env is restored
        self.assertEqual(os.getenv(self.var), self.val)

    def test_remove_dwave(self):
        # set var to val in "global" env
        var = 'DWAVE_TEST'
        val = 'test'
        os.environ[var] = val

        # ensure isolated env removes dwave var
        with isolated_environ(remove_dwave=True):
            self.assertEqual(os.getenv(var), None)

        # ensure "global" env is restored
        self.assertEqual(os.getenv(var), val)

    def test_empty(self):
        # set var to val in "global" env
        self.setval()
        self.assertEqual(os.getenv(self.var), self.val)

        # ensure isolated env starts empty
        kwargs = dict(empty=True)

        def test():
            self.assertEqual(os.getenv(self.var), None)

        with isolated_environ(**kwargs):
            test()

        isolated_environ(**kwargs)(test)()

        # ensure "global" env is restored
        self.assertEqual(os.getenv(self.var), self.val)


class TestMockUtils(unittest.TestCase):

    def test_iterable_mock_open(self):
        data = '1\n2\n3'
        namespaced_open = '{}.open'.format(__name__)

        # first, verify `mock.mock_open` fails for iteration
        with self.assertRaises(AttributeError):
            with mock.patch(namespaced_open, mock.mock_open(data), create=True):
                for line in open('filename'):
                    pass

        # then verify our `iterable_mock_open` works
        lines = []
        with mock.patch(namespaced_open, iterable_mock_open(data), create=True):
            for line in open('filename'):
                lines.append(line)
        self.assertEqual(''.join(lines), data)


class TestSolverDataMocks(unittest.TestCase):

    def test_solver_configuration_data(self):
        data = mocks.solver_configuration_data()
        self.assertGreater(len(data['id']), 1)
        self.assertGreater(len(data['status']), 1)
        self.assertGreater(len(data['description']), 1)
        self.assertIsInstance(data['properties'], dict)
        self.assertIsInstance(data['avg_load'], float)

    def test_structured_solver_data(self):
        data = mocks.structured_solver_data()
        self.assertEqual(data['properties']['category'], 'qpu')
        self.assertIn('topology', data['properties'])
        self.assertIsInstance(data['properties']['qubits'], list)
        self.assertIsInstance(data['properties']['qubits'], list)
        self.assertGreater(len(data['properties']['qubits']), 0)

    def test_qpu_clique_solver_data(self):
        n = 3
        data = mocks.qpu_clique_solver_data(n)
        self.assertEqual(data['id'], 'clique_{}q_mock'.format(n))
        self.assertEqual(data['properties']['num_qubits'], n)
        self.assertEqual(data['properties']['qubits'], list(range(n)))

    @unittest.skipUnless(dimod, "dimod not installed")
    def test_qpu_chimera_solver_data(self):
        # 2 x 2 chimera tiles of 1-1 bipartite graphs, overall forming a cycle over 8 qubits
        m, n, t = 2, 2, 1
        num_qubits = m * n * 2 * t
        data = mocks.qpu_chimera_solver_data(m, n, t)
        nodes = data['properties']['qubits']
        edges = data['properties']['couplers']
        self.assertEqual(data['id'], f'chimera_{num_qubits}q_mock')
        self.assertEqual(data['properties']['num_qubits'], num_qubits)
        self.assertEqual(set(nodes), set(range(num_qubits)))
        self.assertEqual(len(nx.find_cycle(nx.Graph(edges))), num_qubits)

    @unittest.skipUnless(dimod, "dimod not installed")
    def test_qpu_pegasus_solver_data(self):
        m = 2
        num_qubits = 24 * m * (m-1)     # includes non-fabric qubits
        num_edges = 12 * (15 * (m-1)^2 + m - 3)
        data = mocks.qpu_pegasus_solver_data(m)
        nodes = data['properties']['qubits']
        edges = data['properties']['couplers']
        self.assertEqual(data['id'], f'pegasus_{num_qubits}q_mock')
        self.assertEqual(data['properties']['num_qubits'], num_qubits)
        self.assertLessEqual(len(nodes), num_qubits)
        self.assertLessEqual(len(edges), num_edges)

    def test_unstructured_solver_data(self):
        data = mocks.unstructured_solver_data()
        self.assertEqual(data['properties']['category'], 'hybrid')
        self.assertIn('minimum_time_limit', data['properties'])
        self.assertListEqual(data['properties']['supported_problem_types'], ['bqm'])

        data = mocks.hybrid_bqm_solver_data()
        self.assertListEqual(data['properties']['supported_problem_types'], ['bqm'])

        data = mocks.hybrid_dqm_solver_data()
        self.assertListEqual(data['properties']['supported_problem_types'], ['dqm'])


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import copy
import networkx as nx
from parameterized import parameterized, parameterized_class

try:
    import dimod
except ImportError:
    dimod = None

from dwave.cloud.client import Client
from dwave.cloud.solver import StructuredSolver
from dwave.cloud.testing import mocks

try:
    C16 = StructuredSolver(data=mocks.qpu_chimera_solver_data(16), client=None)
    P16 = StructuredSolver(data=mocks.qpu_pegasus_solver_data(16), client=None)
    solvers = [(C16,), (P16,)]
except RuntimeError:
    solvers = []

from tests import config

@parameterized_class(("solver", ), solvers)
@unittest.skipUnless(dimod, "dimod not installed")
class TestCheckProblem(unittest.TestCase):

    def test_identity(self):
        # NOTE: cast to list can be removed once we drop support for dimod 0.8.x
        bqm = dimod.generators.ran_r(1, list(self.solver.edges))
        self.assertTrue(self.solver.check_problem(bqm.linear, bqm.quadratic))

    def test_valid_subgraph(self):
        bqm = dimod.generators.ran_r(1, list(self.solver.edges)[:10])
        self.assertTrue(self.solver.check_problem(bqm.linear, bqm.quadratic))

    def test_invalid_subgraph(self):
        bqm = dimod.generators.ran_r(1, 10)
        self.assertFalse(self.solver.check_problem(bqm.linear, bqm.quadratic))

    def test_invalid_small(self):
        bqm = dimod.BQM.from_ising({}, {'ab': 1, 'bc': 1, 'ca': 1})
        self.assertFalse(self.solver.check_problem(bqm.linear, bqm.quadratic))

    def test_supergraph(self):
        n = max(self.solver.nodes)
        edges = list(self.solver.edges) + [(n+i, n+2*i) for i in range(1, 10)]
        bqm = dimod.generators.ran_r(1, edges)
        self.assertFalse(self.solver.check_problem(bqm.linear, bqm.quadratic))

    def test_legacy_format(self):
        # NOTE: cast to list can be removed once we drop support for dimod 0.8.x
        bqm = dimod.generators.ran_r(1, list(self.solver.edges))
        h = list(bqm.linear.values())   # h as list is still supported
        self.assertTrue(self.solver.check_problem(h, bqm.quadratic))


class TestReformatParameters(unittest.TestCase):
    def test_empty(self):
        self.assertEqual(StructuredSolver.reformat_parameters('ising', {}, {}), {})

    def test_initial_states(self):
        doc = {'initial_state': {0: 0, 4: 1}}

        self.assertEqual(StructuredSolver.reformat_parameters('ising', doc, dict(num_qubits=9)),
                         dict(initial_state=[-1, 3, 3, 3, 1, 3, 3, 3, 3]))
        self.assertEqual(StructuredSolver.reformat_parameters('qubo', doc, dict(num_qubits=9)),
                         dict(initial_state=[0, 3, 3, 3, 1, 3, 3, 3, 3]))

        if dimod:
            self.assertEqual(StructuredSolver.reformat_parameters('SPIN', doc, dict(num_qubits=9)),
                             dict(initial_state=[-1, 3, 3, 3, 1, 3, 3, 3, 3]))
            self.assertEqual(StructuredSolver.reformat_parameters('BINARY', doc, dict(num_qubits=9)),
                             dict(initial_state=[0, 3, 3, 3, 1, 3, 3, 3, 3]))

        self.assertEqual(doc, {'initial_state': {0: 0, 4: 1}})

    def test_initial_states_inplace(self):
        doc = {'initial_state': {0: 0, 4: 1}}
        StructuredSolver.reformat_parameters('ising', doc, dict(num_qubits=9), inplace=True)
        self.assertEqual(doc, dict(initial_state=[-1, 3, 3, 3, 1, 3, 3, 3, 3]))

    def test_initial_states_sequence(self):
        doc = {'initial_state': [-1, 3, 3, 3, 1, 3, 3, 3, 3]}
        self.assertEqual(StructuredSolver.reformat_parameters('ising', doc, dict(num_qubits=9)),
                         dict(initial_state=[-1, 3, 3, 3, 1, 3, 3, 3, 3]))

    def test_vartype_smoke(self):
        for vt in StructuredSolver._handled_problem_types:
            StructuredSolver.reformat_parameters(vt, {}, {})

        with self.assertRaises(ValueError):
            StructuredSolver.reformat_parameters('not a type', {}, {})

    @unittest.skipUnless(dimod, "dimod not installed")
    def test_vartype_dimod_smoke(self):
        StructuredSolver.reformat_parameters('SPIN', {}, {})
        StructuredSolver.reformat_parameters('BINARY', {}, {})
        StructuredSolver.reformat_parameters(dimod.BINARY, {}, {})
        StructuredSolver.reformat_parameters(dimod.SPIN, {}, {})

        with self.assertRaises(ValueError):
            StructuredSolver.reformat_parameters("INTEGER", {}, {})

class QpuAccessTimeEstimate(unittest.TestCase):
    """Test QPU access time estimation method."""

    @classmethod
    def setUpClass(cls):
        try:
            cls.solver_mock = StructuredSolver(data=mocks.qpu_pegasus_solver_data(16,
                problem_timing_data=mocks.qpu_problem_timing_data(qpu='advantage')), client=None)
        except RuntimeError:
            raise cls.skipTest(cls, "structured solver mock not available")

    def test_mock_conflicting_params(self):

        # not allowed annealing_time together with anneal_schedule
        with self.assertRaises(ValueError):
            self.solver_mock.estimate_qpu_access_time(num_qubits=1000,
                anneal_schedule=[[0.0, 0.0], [50.0, 0.5]],
                annealing_time=150)
        # reverse anneal without required initial state
        with self.assertRaises(ValueError):
            self.solver_mock.estimate_qpu_access_time(num_qubits=1000,
                anneal_schedule=[[0.0, 1.0], [2.75, 0.45], [82.75, 0.45], [82.761, 1.0]])
    def test_mock_version(self):
        # currently support is for version 1.0.x
        with self.assertRaises(ValueError):
            solver_mock_ver = copy.deepcopy(self.solver_mock)
            solver_mock_ver.properties["problem_timing_data"]["version"] = '1.1.0'
            solver_mock_ver.estimate_qpu_access_time(num_qubits=1000)

    def test_mock_missing_params(self):
        with self.assertRaises(KeyError):
            solver_mock_ver = copy.deepcopy(self.solver_mock)
            solver_mock_ver.properties["problem_timing_data"].pop("version")
            solver_mock_ver.estimate_qpu_access_time(num_qubits=1000)

    @parameterized.expand([({'num_qubits': 1000}, 15341),
                 ({'num_qubits': 1000, 'num_reads': 500}, 149225),
                 ({'num_qubits': 1000, 'annealing_time': 400}, 15721),
                 ({'num_qubits': 1000, 'anneal_schedule': [[0.0, 0.0], [800.0, 0.5]]}, 16121),
                 ({'num_qubits': 1000, 'anneal_schedule': [[0.0, 1.0], [2, 0.45], [102, 0.45], [102, 1.0]], 'initial_state': "dummy"}, 15427),])
    def test_mock_estimations(self, d, t):
        self.assertEqual(int(self.solver_mock.estimate_qpu_access_time(**d)), t)

    def test_mock_estimations_increase_with_reads(self):
        # increase number of reads
        runtime_r500 = self.solver_mock.estimate_qpu_access_time(num_qubits=1000, num_reads=500)
        runtime_r600 = self.solver_mock.estimate_qpu_access_time(num_qubits=1000, num_reads=600)
        self.assertGreater(runtime_r600, runtime_r500)

    @unittest.skipUnless(config, "No live server configuration available.")
    def test_live_qpu_access_time_estimate(self):
        with Client(**config) as client:
            solver = client.get_solver(topology__type='pegasus')
            try:
                self.assertEqual(solver.name, 'test')
            except:
                print(solver.name)

            # test live qpu is in a somewhat reasonable range
            # results for num_qubits=1000: Advanatge4.1 15341, DW_2000Q_6 11775
            runtime_q1000 = solver.estimate_qpu_access_time(num_qubits=1000)
            self.assertGreater(runtime_q1000, 5000)
            self.assertGreater(20000, runtime_q1000)

            # compare for most qubits in use (best model accuracy in version 1.0.0)
            h = {node: 0.5 for node in list(solver.nodes)[:5000]}
            computation = solver.sample_ising(h, {}, num_reads=1000)
            result = computation.result()
            qpu_access_time = result["timing"]["qpu_access_time"]
            runtime_q5000_r1000 = solver.estimate_qpu_access_time(num_qubits=5000, num_reads=1000)
            miss = abs(qpu_access_time - runtime_q5000_r1000)/max(qpu_access_time, runtime_q5000_r1000)
            self.assertLess(miss, 0.05)



# Copyright 2021 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import requests


class RequestError(requests.exceptions.RequestException):
    """Generic SAPI request error"""

    error_msg = None
    error_code = None

    def __init__(self, *args, **kwargs):
        # exception message populated from, in order of precedence: `args`,
        # `error_msg` kwarg, exception docstring
        self.error_msg = kwargs.pop('error_msg', self.error_msg)
        self.error_code = kwargs.pop('error_code', self.error_code)
        if len(args) < 1 and self.error_msg is not None:
            args = (self.error_msg, )
        if len(args) < 1:
            args = (self.__doc__, )
        super().__init__(*args, **kwargs)


class ResourceBadRequestError(RequestError):
    """Resource failed to parse the request"""

class ResourceAuthenticationError(RequestError):
    """Access to resource not authorized: token is invalid or missing"""

class ResourceAccessForbiddenError(RequestError):
    """Access to resource forbidden"""

class ResourceNotFoundError(RequestError):
    """Resource not found"""

class ResourceConflictError(RequestError):
    """Conflict in the current state of the resource"""

class ResourceLimitsExceededError(RequestError):
    """Number of resource requests exceed the permitted limit"""

class ResourceBadResponseError(RequestError):
    """Unexpected resource response"""

class InternalServerError(RequestError):
    """internal server error occurred while request handling."""

class RequestTimeout(RequestError):
    """API request timed out"""



# Copyright 2021 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import uuid
from typing import Union, Tuple

from dwave.cloud.coders import encode_problem_as_qp, encode_problem_as_ref
from dwave.cloud.utils import utcrel, generate_const_ising_problem
from dwave.cloud.solver import StructuredSolver, UnstructuredSolver
from dwave.cloud.testing.mocks import qpu_clique_solver_data, hybrid_bqm_solver_data


class SapiMockResponses:
    """SAPI mock response generator specialized for a prototypical solver and
    prototypical problem (e.g. structured solver and ising problem).
    """

    def __init__(self,
                 solver: Union[StructuredSolver, UnstructuredSolver],
                 problem: Union[Tuple[dict, dict], 'dimod.BQM'],
                 answer: dict,
                 problem_type: str,
                 problem_id: str = None,
                 problem_label: str = None,
                 ):
        self.solver = solver
        self.problem = problem
        self.answer = answer

        self.solver_id = solver.id
        self.problem_type = problem_type
        self.problem_id = str(uuid.uuid4()) if problem_id is None else problem_id
        self.problem_label = str(uuid.uuid4()) if problem_label is None else problem_label

    def problem_data(self,
                     solver: Union[StructuredSolver, UnstructuredSolver] = None,
                     problem: Union[Tuple[dict, dict], 'dimod.BQM'] = None) -> dict:
        raise NotImplementedError

    def problem_messages(self) -> dict:
        return [{
            "timestamp": utcrel(-10).isoformat(),
            "message": "Internal SAPI error occurred.",
            "severity": "ERROR"
        }]

    def problem_metadata(self, **kwargs) -> dict:
        metadata = {
            "submitted_by":  str(uuid.uuid4()),
            "solver": self.solver_id,
            "type": self.problem_type,
            "submitted_on": utcrel(-20).isoformat(),
            "solved_on": utcrel(-10).isoformat(),
            "status": "COMPLETED",
            "messages": self.problem_messages(),
            "label": self.problem_label,
        }
        metadata.update(**kwargs)
        return metadata

    def problem_answer(self) -> dict:
        return {
            "answer": self.answer.copy()
        }

    def problem_info(self, **kwargs) -> dict:
        response = {
            "id": self.problem_id,
            "data": self.problem_data(),
            "params": {},
            "metadata": self.problem_metadata(),
            "answer": self.answer.copy(),
        }
        response.update(**kwargs)
        return response

    def complete_reply(self, answer_patch=None, **kwargs) -> dict:
        """Reply with solutions for the test problem."""

        response = {
            "status": "COMPLETED",
            "solved_on": utcrel(-10).isoformat(),
            "solver": self.solver_id,
            "submitted_on": utcrel(-20).isoformat(),
            "answer": self.answer.copy(),
            "type": self.problem_type,
            "id": self.problem_id,
            "label": self.problem_label,
        }

        # optional answer fields override
        if answer_patch:
            response['answer'].update(answer_patch)

        # optional top-level override
        response.update(**kwargs)

        return response

    def complete_no_answer_reply(self, **kwargs) -> dict:
        """A reply saying a problem is finished without providing the results."""

        response = {
            "status": "COMPLETED",
            "solved_on": utcrel(-10).isoformat(),
            "solver": self.solver_id,
            "submitted_on": utcrel(-20).isoformat(),
            "type": self.problem_type,
            "id": self.problem_id,
            "label": self.problem_label,
        }
        response.update(**kwargs)
        return response

    def error_reply(self, error_message=None, **kwargs) -> dict:
        """A reply saying an error has occurred."""

        response = {
            "status": "FAILED",
            "solved_on": utcrel(-10).isoformat(),
            "solver": self.solver_id,
            "submitted_on": utcrel(-20).isoformat(),
            "type": self.problem_type,
            "id": self.problem_id,
            "label": self.problem_label,
            "error_message": error_message or 'An error message',
        }
        response.update(**kwargs)
        return response

    def immediate_error_reply(self, code: int, msg: str) -> dict:
        """A reply saying an error has occurred (before scheduling for execution)."""

        return {
            "error_code": code,
            "error_msg": msg
        }

    def cancel_reply(self, **kwargs) -> dict:
        """A reply saying a problem was canceled."""

        return {
            "status": "CANCELLED",
            "solved_on": utcrel(-10).isoformat(),
            "solver": self.solver_id,
            "submitted_on": utcrel(-20).isoformat(),
            "type": self.problem_type,
            "id": self.problem_id,
            "label": self.problem_label,
        }
        response.update(**kwargs)
        return response

    def continue_reply(self, **kwargs) -> dict:
        """A reply saying a problem is still in the queue."""

        response = {
            "status": "PENDING",
            "solved_on": None,
            "solver": self.solver_id,
            "submitted_on": utcrel(0).isoformat(),
            "type": self.problem_type,
            "id": self.problem_id,
            "label": self.problem_label,
        }
        response.update(**kwargs)
        return response



class StructuredSapiMockResponses(SapiMockResponses):

    # TODO: accept problem, generate a consistent answer
    def _problem_answer(self, **kwargs) -> dict:
        answer = {
            "format": "qp",
            "num_variables": 5,
            "energies": "AAAAAAAALsA=",
            "num_occurrences": "ZAAAAA==",
            "active_variables": "AAAAAAEAAAACAAAAAwAAAAQAAAA=",
            "solutions": "AAAAAA==",
            "timing": {}
        }
        answer.update(**kwargs)
        return answer

    def problem_data(self,
                     solver: StructuredSolver = None,
                     problem: Tuple[dict, dict] = None) -> dict:
        if solver is None:
            solver = self.solver
        if problem is None:
            problem = self.problem
        linear, quadratic = problem
        return encode_problem_as_qp(solver, linear, quadratic)

    def __init__(self, **kwargs):
        kwargs.setdefault('solver', StructuredSolver(client=None, data=qpu_clique_solver_data(5)))
        kwargs.setdefault('problem', generate_const_ising_problem(kwargs['solver'], h=1, j=-1))
        kwargs.setdefault('problem_type', 'ising')
        kwargs.setdefault('answer', self._problem_answer())
        super().__init__(**kwargs)


class UnstructuredSapiMockResponses(SapiMockResponses):

    def _problem_answer(self, sampleset, **kwargs) -> dict:
        answer = {
            "format": "bq",
            "data": sampleset.to_serializable()
        }
        answer.update(**kwargs)
        return answer

    def problem_data(self,
                     solver: UnstructuredSolver = None,
                     problem: 'dimod.BQM' = None) -> dict:
        return encode_problem_as_ref(self.problem_data_id)

    def __init__(self, **kwargs):
        import dimod

        solver = UnstructuredSolver(client=None, data=hybrid_bqm_solver_data())
        bqm = dimod.BQM.from_ising({}, {'ab': 1, 'bc': 1, 'ca': 1})
        sampleset = dimod.ExactSolver().sample(bqm)

        kwargs.setdefault('solver', solver)
        kwargs.setdefault('problem', bqm)
        kwargs.setdefault('problem_type', 'bqm')
        kwargs.setdefault('answer', self._problem_answer(sampleset))
        super().__init__(**kwargs)

        # unstructured problem specific
        self.problem_data_id = str(uuid.uuid4())    # mock `self.problem` uploaded



import os
from io import open
from setuptools import setup, find_packages


# Load package info, without importing the package
basedir = os.path.dirname(os.path.abspath(__file__))
package_info_path = os.path.join(basedir, "dwave", "cloud", "package_info.py")
package_info = {}
with open(package_info_path, encoding='utf-8') as f:
    exec(f.read(), package_info)

python_requires = '>=3.7'

# Package requirements, minimal pinning
install_requires = ['requests[socks]>=2.18', 'pydantic>=1.7.3', 'homebase>=1.0',
                    'click>=7.0', 'python-dateutil>=2.7', 'plucky>=0.4.3',
                    'diskcache>=5.2.1']

# Package extras requirements
extras_require = {
    'test': ['requests_mock', 'mock', 'numpy', 'coverage'],

    # bqm support
    'bqm': ['dimod>=0.9.0,<0.13,!=0.10.0,!=0.10.1,!=0.10.2,!=0.10.3,!=0.10.4,!=0.11.4', 'numpy>=1.16'],

    # dqm support
    'dqm': ['dimod>=0.9.7,<0.13,!=0.10.0,!=0.10.1,!=0.10.2,!=0.10.3,!=0.10.4,!=0.11.4', 'numpy>=1.16'],

    # cqm support
    'cqm': ['dimod>=0.10.5,<0.13,!=0.11.4', 'numpy>=1.16'],

    # testing mocks
    'mocks': ['dwave-networkx>=0.8.9'],
}

# Packages provided. Only include packages under the 'dwave' namespace.
# Do not include tests, benchmarks, etc.
packages = [package for package in find_packages() if package.startswith('dwave')]

classifiers = [
    'License :: OSI Approved :: Apache Software License',
    'Operating System :: OS Independent',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
]

setup(
    name=package_info['__packagename__'],
    version=package_info['__version__'],
    author=package_info['__author__'],
    description=package_info['__description__'],
    long_description=open('README.rst', encoding='utf-8').read(),
    url=package_info['__url__'],
    license=package_info['__license__'],
    license_files=["LICENSE"],
    packages=packages,
    python_requires=python_requires,
    install_requires=install_requires,
    extras_require=extras_require,
    classifiers=classifiers,
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'dwave = dwave.cloud.cli:cli'
        ]
    }
)



# Copyright 2017 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Testing utils."""

import os
import functools
from unittest import mock

__all__ = ['mock', 'iterable_mock_open', 'isolated_environ']


def iterable_mock_open(read_data):
    """Version of `mock.mock_open` that supports iteration
    (required when mocking `open` for `configparser.read`).

    Note the difference:

        1) iteration not working with `mock.mock_open`:

            >>> with mock.patch('builtins.open', mock.mock_open('1\n2\n3'), create=True):
            ...  for x in open('asd'):
            ...   print(x)
            ...
            Traceback (most recent call last):
            File "<stdin>", line 1, in <module>
            File "/usr/lib/python3.5/unittest/mock.py", line 2361, in mock_open
                mock.side_effect = reset_data
            AttributeError: 'str' object has no attribute 'side_effect'

        2) iteration working with our `iterable_mock_open`

            >>> with mock.patch('builtins.open', iterable_mock_open('1\n2\n3'), create=True):
            ...  for x in open('asd'):
            ...   print(x)
            ...
            1
            2
            3
    """
    m = mock.mock_open(read_data=read_data)
    m.return_value.__iter__ = lambda self: self
    m.return_value.__next__ = lambda self: next(iter(self.readline, ''))
    return m


class isolated_environ:
    """Context manager and a decorator for modified process environment
    isolation.

    Environment variables can be updated, added and removed. Complete
    environment can be cleared, or cleared only of a subset of variables
    that affect config loading (``DWAVE_*`` vars).

    On context clear, original `os.environ` is restored.

    Args:
        add (dict/Mapping):
            Values to add (or update) to the isolated `os.environ`.

        remove (dict/Mapping, or set/Iterable):
            Values to remove from the isolated `os.environ`.

        remove_dwave (bool, default=False):
            Remove dwave-cloud-client specific variables that affect config
            loading (prefixed with ``DWAVE_`` or ``DW_INTERNAL__``)

        empty (bool, default=False):
            Return empty environment.

    Context:
        Modified copy of global `os.environ`. Restored on context/decorator exit.

    Examples:
        Patch environment in a function scope::

            @isolated_environ(empty=True)
            def f():
                assert len(os.environ) == 0

            f()
            assert len(os.environ) > 0

        Patch environment in a context::

            with isolated_environ(empty=True) as env:
                assert len(os.environ) == 0

            assert len(os.environ) > 0
    """

    def start(self):
        self.patcher = mock.patch.dict(os.environ, values=self.add, clear=self.empty)
        self.patcher.start()

        for key in self.remove:
            os.environ.pop(key, None)

        for key in frozenset(os.environ.keys()):
            if self.remove_dwave and key.startswith("DWAVE_"):
                os.environ.pop(key, None)

        return self

    def stop(self):
        self.patcher.stop()

    def __init__(self, add=None, remove=None, remove_dwave=False, empty=False):
        if add is None:
            add = {}

        if remove is None:
            remove = {}

        self.add = add
        self.remove = remove
        self.remove_dwave = remove_dwave
        self.empty = empty

    def __call__(self, fn):
        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            self.start()
            try:
                return fn(*args, **kwargs)
            finally:
                self.stop()

        return wrapper

    def __enter__(self):
        return self.start()

    def __exit__(self, exc_type, exc_value, traceback):
        self.stop()



# Copyright 2017 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import copy
import base64
import struct
import unittest

try:
    import dimod
except ImportError:
    dimod = None

import numpy as np
from plucky import pluck

from dwave.cloud.coders import (
    encode_problem_as_qp, decode_qp, decode_qp_numpy,
    encode_problem_as_bq, decode_bq, encode_problem_as_ref)
from dwave.cloud.solver import StructuredSolver, UnstructuredSolver
from dwave.cloud.utils import generate_const_ising_problem

# parse string version as tuple
if dimod:
    # python version strings are a lot more permissive, but this covers
    # all the patterns we've used in dimod
    match = re.match('^([0-9]+).([0-9]+).([0-9]+)(?:rc[0-9]*|.dev[0-9]*)*$',
                     dimod.__version__)
    dimod_version = tuple(map(int, match.groups()))
else:
    dimod_version = None


def get_structured_solver():
    data = {
        "properties": {
            "supported_problem_types": ["qubo", "ising"],
            "qubits": [0, 1, 2, 3],
            "couplers": [(0, 1), (1, 2), (2, 3), (3, 0)],
            "num_qubits": 4,
            "parameters": {"num_reads": "Number of samples to return."}
        },
        "id": "test-structured-solver",
        "description": "A test structured solver"
    }
    return StructuredSolver(client=None, data=data)

def get_unstructured_solver():
    data = {
        "properties": {
            "supported_problem_types": ["bqm"],
            "parameters": {"num_reads": "Number of samples to return."}
        },
        "id": "test-unstructured-solver",
        "description": "A test unstructured solver"
    }
    return UnstructuredSolver(client=None, data=data)


class CodersTestBase(unittest.TestCase):
    nan = float('nan')

    # response to a 5-qubit problem from a 5 qubit machine
    res_msg = {
        "id": "test",
        "solver": "test",
        "status": "COMPLETED",
        "type": "ising",
        "answer": {
            "format": "qp",
            "num_variables": 5,
            "energies": "AAAAAAAALsA=",
            "num_occurrences": "ZAAAAA==",
            "active_variables": "AAAAAAEAAAACAAAAAwAAAAQAAAA=",
            "solutions": "AAAAAA==",
            "timing": {}
        }
    }
    res_num_variables = 5
    res_active_variables = (0, 1, 2, 3, 4)
    res_solutions = [[-1, -1, -1, -1, -1]]
    res_energies = (-15.0,)
    res_num_occurrences = (100,)


class TestQPCoders(CodersTestBase):

    def encode_doubles(self, values):
        return base64.b64encode(struct.pack('<' + ('d' * len(values)), *values)).decode('utf-8')

    def test_qp_request_encoding_all_qubits(self):
        """Test biases and coupling strengths are properly encoded (base64 little-endian doubles)."""

        solver = get_structured_solver()
        linear, quadratic = generate_const_ising_problem(solver, h=1, j=-1)
        request = encode_problem_as_qp(solver, linear, quadratic)
        self.assertEqual(request['format'], 'qp')
        self.assertEqual(request['lin'],  self.encode_doubles([1, 1, 1, 1]))
        self.assertEqual(request['quad'], self.encode_doubles([-1, -1, -1, -1]))

    def test_qp_request_encodes_offset(self):
        """Test `offset` is stored in problem data."""

        solver = get_structured_solver()
        linear, quadratic = generate_const_ising_problem(solver, h=1, j=-1)

        # default offset is zero
        request = encode_problem_as_qp(solver, linear, quadratic)
        self.assertEqual(request['offset'], 0)

        # test explicit offset
        offset = 2.0
        request = encode_problem_as_qp(solver, linear, quadratic, offset)
        self.assertEqual(request['offset'], offset)

    def test_qp_request_encoding_sub_qubits(self):
        """Inactive qubits should be encoded as NaNs. Inactive couplers should be omitted."""

        solver = get_structured_solver()
        linear = {0: 1, 1: 1}
        quadratic = {(0, 1): -1}
        request = encode_problem_as_qp(solver, linear, quadratic)
        self.assertEqual(request['format'], 'qp')
        # [1, 1, NaN, NaN]
        self.assertEqual(request['lin'],  self.encode_doubles([1, 1, self.nan, self.nan]))
        # [-1]
        self.assertEqual(request['quad'], self.encode_doubles([-1]))

    def test_qp_request_encoding_missing_qubits(self):
        """Qubits don't have to be specified with biases only, but also with couplings."""

        solver = get_structured_solver()
        linear = {}
        quadratic = {(0, 1): -1}
        request = encode_problem_as_qp(solver, linear, quadratic)
        self.assertEqual(request['format'], 'qp')
        # [0, 0, NaN, NaN]
        self.assertEqual(request['lin'],  self.encode_doubles([0, 0, self.nan, self.nan]))
        # [-1]
        self.assertEqual(request['quad'], self.encode_doubles([-1]))

    def test_qp_request_encoding_sub_qubits_implicit_biases(self):
        """Biases don't have to be specified for qubits to be active."""

        solver = get_structured_solver()
        linear = {}
        quadratic = {(0,3): -1}
        request = encode_problem_as_qp(solver, linear, quadratic)
        self.assertEqual(request['format'], 'qp')
        # [0, NaN, NaN, 0]
        self.assertEqual(request['lin'],  self.encode_doubles([0, self.nan, self.nan, 0]))
        # [-1]
        self.assertEqual(request['quad'], self.encode_doubles([-1]))

    def test_qp_request_encoding_sub_qubits_implicit_couplings(self):
        """Couplings should be zero for active qubits, if not specified."""

        solver = get_structured_solver()
        linear = {0: 0, 3: 0}
        quadratic = {}
        request = encode_problem_as_qp(solver, linear, quadratic)
        self.assertEqual(request['format'], 'qp')
        # [0, NaN, NaN, 0]
        self.assertEqual(request['lin'],  self.encode_doubles([0, self.nan, self.nan, 0]))
        # [0]
        self.assertEqual(request['quad'], self.encode_doubles([0]))

    def test_qp_request_encoding_undirected_biases(self):
        """Quadratic terms are correctly encoded when given as undirected biases."""

        solver = get_structured_solver()
        linear = {}
        quadratic = {(0,3): -1, (3,0): -1}
        request = encode_problem_as_qp(solver, linear, quadratic, undirected_biases=True)
        self.assertEqual(request['format'], 'qp')
        # [0, NaN, NaN, 0]
        self.assertEqual(request['lin'],  self.encode_doubles([0, self.nan, self.nan, 0]))
        # [-1]
        self.assertEqual(request['quad'], self.encode_doubles([-1]))


class TestQPDecoders(CodersTestBase):
    """Test QP decoders correctly decode response data."""

    def test_qp_response_decoding(self):
        res = decode_qp(copy.deepcopy(self.res_msg))

        self.assertEqual(res.get('format'), 'qp')
        self.assertEqual(res.get('num_variables'), self.res_num_variables)
        self.assertEqual(res.get('active_variables'), self.res_active_variables)
        self.assertEqual(res.get('solutions'), self.res_solutions)
        self.assertEqual(res.get('energies'), self.res_energies)
        self.assertEqual(res.get('num_occurrences'), self.res_num_occurrences)

    def test_qp_response_numpy_decoding(self):
        res = decode_qp_numpy(copy.deepcopy(self.res_msg), return_matrix=False)

        self.assertEqual(res.get('format'), 'qp')
        self.assertEqual(res.get('num_variables'), self.res_num_variables)
        self.assertEqual(res.get('active_variables'), list(self.res_active_variables))
        self.assertEqual(res.get('solutions'), list(self.res_solutions))
        self.assertEqual(res.get('energies'), list(self.res_energies))
        self.assertEqual(res.get('num_occurrences'), list(self.res_num_occurrences))

    def test_qp_response_numpy_decoding_numpy_array(self):
        res = decode_qp_numpy(copy.deepcopy(self.res_msg), return_matrix=True)

        self.assertEqual(res.get('format'), 'qp')
        self.assertEqual(res.get('num_variables'), self.res_num_variables)
        np.testing.assert_array_equal(res.get('active_variables'), np.array(self.res_active_variables))
        np.testing.assert_array_equal(res.get('solutions'), np.array(self.res_solutions))
        np.testing.assert_array_equal(res.get('energies'), np.array(self.res_energies))
        np.testing.assert_array_equal(res.get('num_occurrences'), np.array(self.res_num_occurrences))


class TestQPDecodersWithOffset(TestQPDecoders):
    """Test decoders correctly apply energy offset."""

    res_offset = 2.0

    def setUp(self):
        self.res_msg['answer']['offset'] = self.res_offset
        self.res_energies = [en + self.res_offset for en in self.res_energies]

    def test_offset_present_in_answer(self):
        res = decode_qp(copy.deepcopy(self.res_msg))
        self.assertEqual(res.get('offset'), self.res_offset)


@unittest.skipUnless(dimod, "dimod required for BQ coders")
class TestBQCoders(unittest.TestCase):

    @unittest.skipUnless(dimod and dimod_version < (0, 10, 0), "dimod < 0.10 required for BQ encoder")
    def test_bq_encodes_empty_bqm(self):
        """Empty BQM has to be trivially encoded."""

        bqm = dimod.BQM.from_qubo({})
        req = encode_problem_as_bq(bqm)

        self.assertEqual(req.get('format'), 'bq')
        self.assertEqual(pluck(req, 'data.version.bqm_schema'), '3.0.0')
        self.assertEqual(pluck(req, 'data.num_variables'), 0)
        self.assertEqual(pluck(req, 'data.num_interactions'), 0)

    @unittest.skipUnless(dimod and dimod_version < (0, 10, 0), "dimod < 0.10 required for BQ encoder")
    def test_bq_encodes_ising_bqm(self):
        """Simple Ising BQM properly encoded."""

        bqm = dimod.BQM.from_ising({0: 1}, {(0, 1): 1})
        req = encode_problem_as_bq(bqm)

        self.assertEqual(req.get('format'), 'bq')
        self.assertEqual(pluck(req, 'data.version.bqm_schema'), '3.0.0')
        self.assertEqual(pluck(req, 'data.variable_type'), 'SPIN')
        self.assertEqual(pluck(req, 'data.num_variables'), 2)
        self.assertEqual(pluck(req, 'data.num_interactions'), 1)

    @unittest.skipUnless(dimod and dimod_version < (0, 10, 0), "dimod < 0.10 required for BQ encoder")
    def test_bq_encodes_qubo_bqm(self):
        """Simple Qubo BQM properly encoded."""

        bqm = dimod.BQM.from_qubo({(0, 1): 1})
        req = encode_problem_as_bq(bqm)

        self.assertEqual(req.get('format'), 'bq')
        self.assertEqual(pluck(req, 'data.version.bqm_schema'), '3.0.0')
        self.assertEqual(pluck(req, 'data.variable_type'), 'BINARY')
        self.assertEqual(pluck(req, 'data.num_variables'), 2)
        self.assertEqual(pluck(req, 'data.num_interactions'), 1)

    @unittest.skipUnless(dimod and dimod_version < (0, 10, 0), "dimod < 0.10 required for BQ encoder")
    def test_bq_encodes_bqm_with_named_vars(self):
        """BQM with named variable properly encoded."""

        bqm = dimod.BQM.from_ising({}, {'ab': 1, 'bc': 1, 'ca': 1})
        req = encode_problem_as_bq(bqm)

        self.assertEqual(req.get('format'), 'bq')
        self.assertEqual(pluck(req, 'data.version.bqm_schema'), '3.0.0')
        self.assertEqual(pluck(req, 'data.variable_type'), 'SPIN')
        self.assertEqual(pluck(req, 'data.num_variables'), 3)
        self.assertEqual(pluck(req, 'data.num_interactions'), 3)
        self.assertEqual(pluck(req, 'data.variable_labels'), list('abc'))

    def test_ref_encoder(self):
        problem_id = '123'

        req = encode_problem_as_ref(problem_id)

        self.assertEqual(req.get('format'), 'ref')
        self.assertEqual(req.get('data'), problem_id)

    def test_bq_ref_input_validation(self):
        problem_id = '123'
        with self.assertRaises(TypeError):
            encode_problem_as_bq(problem_id)

        bqm = dimod.BQM.from_qubo({})
        with self.assertRaises(TypeError):
            encode_problem_as_ref(bqm)

    def test_bq_response_decoding(self):
        """Answer to simple problem properly decoded."""

        problem_type = 'bqm'
        ss = dimod.SampleSet.from_samples(
            ([[0, 1], [1, 0]], 'ab'), vartype='BINARY', energy=0)

        msg = dict(
            answer=dict(format='bq', data=ss.to_serializable()),
            type=problem_type)

        res = decode_bq(msg)

        self.assertEqual(res.get('problem_type'), problem_type)
        self.assertEqual(res.get('sampleset'), ss)



# Copyright 2017 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dwave.cloud.api import exceptions as _api


# alias for backward compatibility
SAPIError = _api.RequestError


class ConfigFileError(Exception):
    """Base exception for all config file processing errors."""

class ConfigFileReadError(ConfigFileError):
    """Non-existing or unreadable config file specified or implied."""

class ConfigFileParseError(ConfigFileError):
    """Invalid format of config file."""


class SolverError(_api.RequestError):
    """Generic solver-related error"""

class ProblemError(_api.RequestError):
    """Generic problem-related error"""


class SolverFailureError(SolverError):
    """Remote failure calling a solver"""

class SolverNotFoundError(_api.ResourceNotFoundError, SolverError):
    """Solver with matching feature set not found / not available"""

class ProblemNotFoundError(_api.ResourceNotFoundError, ProblemError):
    """Problem not found"""

class SolverOfflineError(SolverError):
    """Action attempted on an offline solver"""

class SolverAuthenticationError(_api.ResourceAuthenticationError, SolverError):
    """Invalid token or access denied"""

class UnsupportedSolverError(SolverError):
    """The solver received from the API is not supported by the client"""

class SolverPropertyMissingError(UnsupportedSolverError):
    """The solver received from the API does not have required properties"""


class Timeout(_api.RequestError):
    """Deprecated and unused."""

class PollingTimeout(Exception):
    """Problem polling timed out."""

# for backward compatibility
from dwave.cloud.api.exceptions import RequestTimeout


class CanceledFutureError(Exception):
    """An exception raised when code tries to read from a canceled future."""

    def __init__(self):
        super().__init__("An error occurred reading results from a canceled request")


class InvalidAPIResponseError(_api.ResourceBadResponseError):
    """Unexpected response from D-Wave Solver API"""


class InvalidProblemError(ValueError):
    """Solver cannot handle the given binary quadratic model."""

class ProblemStructureError(InvalidProblemError):   # inherit for backwards compat
    """Problem structure incompatible with a structured solver graph."""


class ProblemUploadError(Exception):
    """Problem multipart upload failed."""
